// g++ -o mp2osc mp2osc.cpp -O2 -Wall -lcurl -pthread

#include <cstring>
#include <string>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdbool.h>
#include <pthread.h>
#include "curl/curl.h"

using namespace std;

const char * backend_url = "http://osm.svimik.com/photosync/workers_backend.php?task=mp2osc";
const char * mapillary_client_id; // argv[1]
const char * mapillary_key; // argv[2]
const char * mapillary_sequence; // argv[3]
const char * openstreetcam_key; // argv[4]
const char * task_id; // argv[5]
char openstreetcam_sequence_id[32] = {0}; // obtained through API call

int progress_images_discovered = 0;
int progress_images_downloaded = 0;
int progress_images_uploaded = 0;
int progress_dw_speed = 0;
int progress_ul_speed = 0;
long progress_bytes_uploaded = 0;

char last_error[1024] = {0}; // threads will store the last error description here
FILE * logfile = NULL;
pthread_mutex_t printf_mutex;
curl_version_info_data * libcurl_version;

bool date(const char * format, char * dst, int dst_len, long ts){
	time_t t = (time_t)(ts/1000);
	struct tm lt;
	gmtime_r(&t, &lt);
	return strftime(dst, dst_len, format, &lt);
}

typedef struct{
	char key[128];
	char heading[16];
	char lon[16];
	char lat[16];
	int width;
	int height;
	char captured_at[32];
	char download_url[256];
	bool downloaded;
	bool uploaded;
	string data;
} image;
image * images = NULL;

int pthread_create_detached(pthread_t *thread, void *(*start_routine) (void *), void *arg){
	pthread_attr_t tattr;
	pthread_attr_init(&tattr);
	pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
	return pthread_create(thread, &tattr, start_routine, arg);
}

const char * strstr_right(const char * str1, const char * str2){
	const char * pos = strstr(str1, str2);
	if(!pos){return NULL;}
	return pos + strlen(str2);
}

double microtime(){
	struct timespec tp;
	clock_gettime(CLOCK_MONOTONIC, &tp);
	return (double)tp.tv_sec + (double)tp.tv_nsec / 1000000000.00;
}

bool file_exists(const char * fn){
	return (access(fn, F_OK) == 0) ? true : false;
}

void log(const char* format, ...){
	// get time
	time_t timer;
	time(&timer);
	struct tm * tm_info = localtime(&timer);
	// print
	pthread_mutex_lock(&printf_mutex);
	printf("[%02i:%02i:%02i] ", tm_info->tm_hour, tm_info->tm_min, tm_info->tm_sec);
	va_list args;
	va_start(args, format);
	vprintf(format, args);
	va_end(args);
	if(logfile){
		fprintf(logfile, "[%02i:%02i:%02i] ", tm_info->tm_hour, tm_info->tm_min, tm_info->tm_sec);
		va_start(args, format);
		vfprintf(logfile, format, args);
		va_end(args);
		fflush(logfile);
	}
	pthread_mutex_unlock(&printf_mutex);
}

/* Some generic curl callback */

size_t curl_callback(char *ptr, size_t size, size_t nmemb, string *str){
	str->append(ptr, size*nmemb);
	//log("Curl got %ld bytes\r\n", size*nmemb);
	return size*nmemb;
}

/* To make the POST requests easier */

CURLcode curl_post(CURL *curl, const char * url, const char* format, ...){
	char post_data[1024];
	va_list args;
	va_start(args, format);
	vsnprintf(post_data, sizeof(post_data), format, args);
	va_end(args);
	curl_easy_setopt(curl, CURLOPT_URL, url);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_data);
	return curl_easy_perform(curl);
}

/* Different curl options for different backends */

CURL * curl_init_backend(){
	CURL * curl = curl_easy_init();
	if(!curl){
		return NULL;
	}
	// curl bug workaround
	if(libcurl_version->version_num < 0x073200){
		curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
	}
	// additional headers
	struct curl_slist * headers = NULL;
	headers = curl_slist_append(headers, "Content-Type: application/x-www-form-urlencoded; charset=UTF-8");
	headers = curl_slist_append(headers, "Connection: keep-alive");
	// curl options
	curl_easy_setopt(curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
	curl_easy_setopt(curl, CURLOPT_POST, 1);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_callback);
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 10);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, 30);
	//curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);
	return curl;
}

CURL * curl_init_mapillary(){
	CURL * curl = curl_easy_init();
	if(!curl){
		return NULL;
	}
	// curl bug workaround
	if(libcurl_version->version_num < 0x073200){
		curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
	}
	// additional headers
	char auth_header[1024];
	snprintf(auth_header, sizeof(auth_header), "Authorization: Bearer %s", mapillary_key);
	struct curl_slist * headers = NULL;
	headers = curl_slist_append(headers, "Content-Type: application/x-www-form-urlencoded; charset=UTF-8");
	headers = curl_slist_append(headers, "Connection: keep-alive");
	headers = curl_slist_append(headers, auth_header);
	// curl options
	curl_easy_setopt(curl, CURLOPT_ENCODING, "gzip,deflate");
	curl_easy_setopt(curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_callback);
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 30);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, 120);
	return curl;
}

CURL * curl_init_downloader(){
	CURL * curl = curl_easy_init();
	if(!curl){
		return NULL;
	}
	// curl bug workaround
	if(libcurl_version->version_num < 0x073200){
		curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
	}
	// additional headers
	struct curl_slist * headers = NULL;
	headers = curl_slist_append(headers, "Connection: keep-alive");
	headers = curl_slist_append(headers, "Accept:");
	// curl options
	curl_easy_setopt(curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_callback);
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 30);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, 120);
	return curl;
}

CURL * curl_init_openstreetcam(){
	CURL * curl = curl_easy_init();
	if(!curl){
		return NULL;
	}
	// curl bug workaround
	if(libcurl_version->version_num < 0x073200){
		curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
	}
	// additional headers
	struct curl_slist * headers=NULL;
	headers = curl_slist_append(headers, "Connection: keep-alive");
	headers = curl_slist_append(headers, "Expect:");
	// curl options
	curl_easy_setopt(curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_callback);
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 30);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, 300);
	curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0); // yes, they have weird cert, whatever
	//curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);
	return curl;
}

/* Report the current progress */

bool progress(CURL * curl, bool finished, bool success, const char * errstr){
	string data = "";
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &data);
	char progress[1024];
	snprintf(progress, sizeof(progress), "{\"images_discovered\":%i,\"images_downloaded\":%i,\"images_uploaded\":%i,\"dw_speed\":%i,\"ul_speed\":%i,\"bytes_uploaded\":%li,\"osc_sequence_id\":\"%s\",\"errstr\":\"%s\"}", progress_images_discovered, progress_images_downloaded, progress_images_uploaded, progress_dw_speed, progress_ul_speed, progress_bytes_uploaded, openstreetcam_sequence_id, (errstr ? errstr : ""));
	CURLcode res = curl_post(curl, backend_url, "task_id=%s&progress=%s&finished=%i&success=%i", task_id, progress, (finished ? 1 : 0), (success ? 1 : 0));
	if(res != CURLE_OK){
		log("[progress] Curl returned: %s\r\n", curl_easy_strerror(res));
		return false;
	}
	long response_code;
	curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &response_code);
	if(response_code != 200){
		log("[progress] Server returned code: %li\r\n", response_code);
		return false;
	}
	if(strcmp(data.c_str(), "OK") != 0){
		log("[progress] Server returned: %s\r\n", data.c_str());
		return false;
	}
	log("[progress] OK\r\n");
	return true;
}

/* Get the list of images for a given sequence */

bool get_mapillary_image_list(CURL * curl){
	char url[256];
	snprintf(url, sizeof(url), "https://a.mapillary.com/v3/sequences?client_id=%s&sequence_keys=%s", mapillary_client_id, mapillary_sequence);
	string data = "";
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &data);
	curl_easy_setopt(curl, CURLOPT_POST, 0);
	curl_easy_setopt(curl, CURLOPT_URL, url);
	CURLcode res = curl_easy_perform(curl);
	if(res != CURLE_OK){
		log("[sequence] Curl returned: %s\r\n", curl_easy_strerror(res));
		snprintf(last_error, sizeof(last_error), "Failed to get image list, curl error %i", res);
		return false;
	}
	long response_code;
	curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &response_code);
	if(response_code != 200){
		log("[sequence] Server returned code: %li\r\n", response_code);
		log("[sequence] Server said: %.1024s\r\n", data.c_str());
		snprintf(last_error, sizeof(last_error), "Failed to get image list, server returned code %li", response_code);
		return false;
	}

	const char * json = data.c_str();
	const char * cas = strstr_right(json, "\"cas\":[");
	const char * image_keys = strstr_right(json, "\"image_keys\":[");
	const char * coordinates = strstr_right(json, "\"coordinates\":[");
	if(!cas || !image_keys || !coordinates){
		log("[sequence] Invalid json returned\r\n");
		log("%s\r\n", data.c_str());
		snprintf(last_error, sizeof(last_error), "Failed to get image list, invalid json returned");
		return false;
	}
	// count number of array elements
	int num = 0;
	int ptr = 0;
	while(cas[ptr]){
		if(cas[ptr] == ','){
			num++;
		}else if(cas[ptr] == ']'){
			if(ptr != 0){
				num++;
			}
			break;
		}
		ptr++;
	}
	log("[sequence] %i images found\r\n", num);
	if(num == 0){
		snprintf(last_error, sizeof(last_error), "Failed to get image list, empty list returned");
		return false;
	}else{
		progress_images_discovered = num;
	}
	// allocate the image array
	images = (image *)calloc(num, sizeof(image));
	if(!images){
		log("[sequence] Out of memory\r\n");
		snprintf(last_error, sizeof(last_error), "Failed to get image list, out of memory");
		return false;
	}
	// parse headings
	int k = 0;
	size_t wptr = 0;
	ptr = 0;
	while(cas[ptr]){
		if(k >= progress_images_discovered){
			break; // should not happen
		}
		char c = cas[ptr++];
		if(c == ','){
			k++;
			wptr = 0;
			continue;
		}else if(c == ']'){
			break;
		}else if(c != '"' && wptr < sizeof(images[k].heading) - 1){
			images[k].heading[wptr++] = c;
		}
	}
	if(k != progress_images_discovered - 1){
		log("[sequence] Parse failure (cas)\r\n");
		snprintf(last_error, sizeof(last_error), "Failed to get image list, parser failed (cas)");
		return false;
	}
	// parse keys
	k = 0;
	wptr = 0;
	ptr = 0;
	while(image_keys[ptr]){
		if(k >= progress_images_discovered){
			break; // should not happen
		}
		char c = image_keys[ptr++];
		if(c == ','){
			k++;
			wptr = 0;
			continue;
		}else if(c == ']'){
			break;
		}else if(c != '"' && wptr < sizeof(images[k].key) - 1){
			images[k].key[wptr++] = c;
		}
	}
	if(k != progress_images_discovered - 1){
		log("[sequence] Parse failure (image_keys)\r\n");
		snprintf(last_error, sizeof(last_error), "Failed to get image list, parser failed (image_keys)");
		return false;
	}
	// parse lon, lat
	// "coordinates":[[24.8556035,59.432635],[24.8556894,59.4326472],[24.8557751,59.4326592],	]
	// "coordinates":[[11.5477891,48.0862444],[11.5477891,48.0862444]] // twice for 1-picture sequence!
	k = 0;
	wptr = 0;
	ptr = 0;
	int sk = 0;
	while(coordinates[ptr]){
		if(k >= progress_images_discovered){
			break; // should not happen
		}
		char prev_c = coordinates[ptr-1];
		char c = coordinates[ptr++];
		if(prev_c == ']' && c == ','){
			k++;
			continue;
		}else if(prev_c == ']' && c == ']'){
			break;
		}else if(c == '[' || c == ']'){
			sk = 0;
			wptr = 0;
		}else if(c == ','){
			sk = 1;
			wptr = 0;
		}else if(sk == 0 && c != '"' && wptr < sizeof(images[k].lon) - 1){
			images[k].lon[wptr++] = c;
		}else if(sk == 1 && c != '"' && wptr < sizeof(images[k].lat) - 1){
			images[k].lat[wptr++] = c;
		}
	}
	if(k == 1 && progress_images_discovered == 1){
		k = 0; // workaround because API may return two coordinates for 1-picture sequence
	}
	if(k != progress_images_discovered - 1){
		log("%s\r\n", json);
		log("[sequence] Parse failure (coordinates), k=%i, expected=%i\r\n", k, progress_images_discovered - 1);
		snprintf(last_error, sizeof(last_error), "Failed to get image list, parser failed (coordinates)");
		return false;
	}
	// init the strings
	for(k=0; k<progress_images_discovered; k++){
		new(&images[k].data) string();
	}
	// debug
	//for(k=0; k<progress_images_discovered; k++){
	//	log("%s %s %s %s\r\n", images[k].key, images[k].heading, images[k].lon, images[k].lat);
	//}
	log("[sequence] Parsed\r\n");
	return true;
}

bool mapillary_parse_metadata(const char * buf, size_t len, int k){
	const char * end = buf + len;
	const char * fields[] = {"original_url", "captured_at", "width", "height"};
	int i;
	for(i=0; i<4; i++){
		char search[256];
		snprintf(search, sizeof(search), "\"%s\":{\"$type\":\"atom\",\"value\":", fields[i]);
		const char * tmp = strstr_right(buf, search);
		if(!tmp || tmp > end){
			log("[links] Failed to find %s field for image %s\r\n", fields[i], images[k].key);
			snprintf(last_error, sizeof(last_error), "Failed to parse image metadata, can't find %s field for image %s", fields[i], images[k].key);
			return false;
		}
		const char * f_end;
		if(tmp[0] == '"'){
			tmp++;
			f_end = strchr(tmp, '"');
		}else{
			f_end = strchr(tmp, '}');
		}
		if(!f_end || f_end > end){
			log("[links] Parser failure on field %s for image %s\r\n", fields[i], images[k].key);
			snprintf(last_error, sizeof(last_error), "Failed to parse image metadata, parser failure on field %s field for image %s", fields[i], images[k].key);
			return false;
		}
		size_t f_len = f_end - tmp;
		if(i == 0){
			if(f_len >= sizeof(images[k].download_url)){
				log("[links] Too long URL (%i)\r\n", fields[i], images[k].key, f_len);
				snprintf(last_error, sizeof(last_error), "Failed to get download link, too long url");
				return false;
			}
			strncpy(images[k].download_url, tmp, f_len);
		}else if(i == 1){
			long ts = atol(tmp);
			if(ts < 1 || !date("%Y-%m-%d %H:%M:%S", images[k].captured_at, sizeof(images[k].captured_at), ts)){
				log("[links] Failed to parse timestamp\r\n", fields[i], images[k].key);
				snprintf(last_error, sizeof(last_error), "Failed to parse timestamp");
				return false;
			}
		}else if(i == 2){
			images[k].width = atoi(tmp);
		}else if(i == 3){
			images[k].height = atoi(tmp);
		}
	}
	return true;
}

bool mapillary_get_download_links(CURL * curl, int item_limit){
	// build the request url
	int limit = item_limit;
	char url[2048];
	int len = 0;
	len+=snprintf(url+len, sizeof(url)-len, "https://a.mapillary.com/v3/model.json?client_id=%s&method=get&paths=%s", mapillary_client_id, "%5B");
	int k;
	// start from progress_images_uploaded
	for(k=progress_images_uploaded; k<progress_images_discovered; k++){
		if(images[k].download_url[0] != '\0'){
			continue;
		}
		//len+=snprintf(url+len, sizeof(url)-len, "%s%s%s", "%5B%22imageByKey%22%2C%22", images[k].key, "%22%2C%5B%22original_url%22%5D%5D%2C");
		len+=snprintf(url+len, sizeof(url)-len, "%s%s%s", "%5B%22imageByKey%22%2C%22", images[k].key, "%22%2C%5B%22height%22%2C%22width%22%2C%22original_url%22%2C%22captured_at%22%5D%5D%2C");
		if(--limit == 0){
			break;
		}
	}
	len-=3;
	len+=snprintf(url+len, sizeof(url)-len, "%s", "%5D");
	// make the request
	string data = "";
	int attempts = 3;
	while(1){
		data = "";
		curl_easy_setopt(curl, CURLOPT_WRITEDATA, &data);
		curl_easy_setopt(curl, CURLOPT_POST, 0);
		curl_easy_setopt(curl, CURLOPT_URL, url);
		CURLcode res = curl_easy_perform(curl);
		if(res != CURLE_OK){
			log("[links] Curl returned: %s\r\n", curl_easy_strerror(res));
			if(attempts > 0){
				attempts--;
				sleep(3);
				continue;
			}else{
				snprintf(last_error, sizeof(last_error), "Failed to get download link, curl error %i", res);
				return false;
			}
		}
		long response_code;
		curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &response_code);
		if(response_code != 200){
			log("[links] Server returned code: %li\r\n", response_code);
			log("[links] Server said: %.1024s\r\n", data.c_str());
			if(attempts > 0){
				attempts--;
				sleep(3);
				continue;
			}else{
				snprintf(last_error, sizeof(last_error), "Failed to get download link, server returned code %li", response_code);
				return false;
			}
		}
		break;
	}
	// parse
	const char * json = data.c_str();
	limit = item_limit;
	// start from progress_images_uploaded
	for(k=progress_images_uploaded; k<progress_images_discovered; k++){
		if(images[k].download_url[0] != '\0'){
			continue;
		}
		char search[256];
		snprintf(search, sizeof(search), "\"%s\":{", images[k].key);
		const char * tmp = strstr_right(json, search);
		if(!tmp){
			log("[links] Failed to retrieve image metadata for image %s\r\n", images[k].key);
			log("%s\r\n\r\n%s\r\n\r\n", json, search);
			snprintf(last_error, sizeof(last_error), "Failed to get image metadata, can't find image %s in API response", images[k].key);
			return false;
		}
		const char * end = strstr(tmp, "}}");
		size_t len = end - tmp;
		if(!end || len < 5){
			log("[links] Parse failure\r\n");
			snprintf(last_error, sizeof(last_error), "Failed to get metadata, parser error");
			return false;
		}
		if(!mapillary_parse_metadata(tmp, len, k)){
			log("[links] mapillary_parse_metadata() failed\r\n");
			return false;
		}
		if(--limit == 0){
			break;
		}
	}
	//for(k=0; k<progress_images_discovered; k++){
	//	log("%s %s %s %s %i %i %s %s\r\n", images[k].key, images[k].heading, images[k].lon, images[k].lat, images[k].width, images[k].height, images[k].captured_at, images[k].download_url);
	//}
	return true;
}

bool mapillary_download(CURL * curl, int k){
	images[k].data = "";
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &images[k].data);
	curl_easy_setopt(curl, CURLOPT_POST, 0);
	curl_easy_setopt(curl, CURLOPT_URL, images[k].download_url);
	double t1 = microtime();
	CURLcode res = curl_easy_perform(curl);
	double t2 = microtime();
	if(res != CURLE_OK){
		log("[dl] Curl returned: %s\r\n", curl_easy_strerror(res));
		snprintf(last_error, sizeof(last_error), "Failed to download image %i, curl error %i", k, res);
		return false;
	}
	long response_code;
	curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &response_code);
	if(response_code != 200){
		log("[dl] Server returned code: %li\r\n", response_code);
		log("[dl] Server said: %.1024s\r\n", images[k].data.c_str());
		snprintf(last_error, sizeof(last_error), "Failed to download image %i, server returned code %li", k, response_code);
		return false;
	}
	const char * buf = images[k].data.c_str();
	int len = images[k].data.length();
	if(len < 3 || memcmp(buf, "\xFF\xD8\xFF", 3) != 0){
		log("[dl] Not a jpeg image\r\n");
		snprintf(last_error, sizeof(last_error), "Failed to download image %i, not a jpeg image", k);
		return false;
	}
	progress_dw_speed = (int)((double)len/(t2-t1));
	log("[dl] %i/%i, %i bytes in %.3fs DW speed: %i bytes/s, %ix%i, %s\r\n", k+1, progress_images_discovered, len, (t2-t1), progress_dw_speed, images[k].width, images[k].height, images[k].captured_at);
	return true;
}

bool thread_dl_running = false;
void * thread_dl(void * arg){
	CURL * curl_mapillary = (CURL *)arg;
	CURL * curl = curl_init_downloader();
	if(!curl){
		thread_dl_running = false;
		return NULL;
	}

	// start from progress_images_uploaded
	int k;
	for(k=progress_images_uploaded; k<progress_images_discovered; k++){
		if(images[k].download_url[0] == '\0'){
			// wait until queue is <5
			while(progress_images_uploaded < k - 5){
				usleep(250000);
			}
			// download new portion of links
			int i;
			bool downloaded = false;
			for(i=0; i<5; i++){
				if(mapillary_get_download_links(curl_mapillary, 10)){
					downloaded = true;
					break;
				}
				sleep(3);
			}
			if(!downloaded){
				break;
			}
			if(i > 0){
				last_error[0] = '\0'; // clear the error if recovered
			}
		}
		if(images[k].download_url[0] == '\0'){
			log("[dl] Impossible error\r\n");
			break;
		}
		// download the image
		int i;
		bool downloaded = false;
		for(i=0; i<10; i++){
			if(mapillary_download(curl, k)){
				downloaded = true;
				break;
			}
			sleep(3);
		}
		if(!downloaded){
			break;
		}
		if(i > 0){
			last_error[0] = '\0'; // clear the error if recovered
		}
		images[k].downloaded = true;
		progress_images_downloaded++;
		// dbg
		//FILE * f = fopen("img.jpg", "wb");
		//fwrite(buf, len, 1, f);
		//fclose(f);
		//break;
	}

	log("[dl] Downloading thread finished, %i/%i images downloaded\r\n", progress_images_downloaded, progress_images_discovered);
	curl_easy_cleanup(curl);
	thread_dl_running = false;
	return NULL;
}

bool openstreetcam_new_sequence(CURL * curl){
	string data = "";
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &data);
	bool is_sphere = (images[progress_images_uploaded].height > 0 && images[progress_images_uploaded].height * 2 == images[progress_images_uploaded].width);
	CURLcode res = curl_post(curl, "https://openstreetcam.org/1.0/sequence/", "obdInfo=0&platformName=photosync&platformVersion=%.25s&access_token=%s%s", mapillary_sequence, openstreetcam_key, (is_sphere ? "&projection=SPHERE" : ""));
	if(res != CURLE_OK){
		log("[sequence_create] Curl returned: %s\r\n", curl_easy_strerror(res));
		snprintf(last_error, sizeof(last_error), "Failed to create sequence, curl error %i", res);
		return false;
	}
	long response_code;
	curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &response_code);
	if(response_code != 200){
		log("[sequence_create] Server returned code: %li\r\n", response_code);
		log("[sequence_create] Server said: %.1024s\r\n", data.c_str());
		snprintf(last_error, sizeof(last_error), "Failed to create sequence, server returned code %li", response_code);
		return false;
	}
	log("%s\r\n", data.c_str());
	const char * json = data.c_str();
	const char * tmp = strstr_right(json, "\"sequence\":{\"id\":\"");
	if(!tmp){
		log("[sequence_create] Invalid json returned\r\n");
		snprintf(last_error, sizeof(last_error), "Failed to create sequence, invalid json returned");
		return false;
	}
	const char * end = strchr(tmp, '"');
	size_t len = end - tmp;
	if(!end || len < 1){
		log("[sequence_create] Parse failure\r\n");
		snprintf(last_error, sizeof(last_error), "Failed to create sequence, parser error");
		return false;
	}
	if(len >= sizeof(openstreetcam_sequence_id)){
		log("[links] Too long sequence id\r\n");
		snprintf(last_error, sizeof(last_error), "Failed to create sequence, too long sequence id");
		return false;
	}
	strncpy(openstreetcam_sequence_id, tmp, len);
	log("[sequence_create] Sequence created, id: %s, sphere=%i\r\n", openstreetcam_sequence_id, (is_sphere ? 1 : 0));
	return true;
}

bool openstreetcam_close_sequence(CURL * curl){
	string data = "";
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &data);
	CURLcode res = curl_post(curl, "https://openstreetcam.org/1.0/sequence/finished-uploading/", "sequenceId=%s&access_token=%s", openstreetcam_sequence_id, openstreetcam_key);
	if(res != CURLE_OK){
		log("[sequence_close] Curl returned: %s\r\n", curl_easy_strerror(res));
		snprintf(last_error, sizeof(last_error), "Failed to close sequence, curl error %i", res);
		return false;
	}
	long response_code;
	curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &response_code);
	if(response_code != 200){
		log("[sequence_close] Server returned code: %li\r\n", response_code);
		log("[sequence_close] Server said: %.1024s\r\n", data.c_str());
		snprintf(last_error, sizeof(last_error), "Failed to close sequence, server returned code %li", response_code);
		return false;
	}
	log("[sequence_close] Sequence closed\r\n");
	return true;
}

bool openstreetcam_upload(CURL * curl, int k){
	char photo_index[32];
	snprintf(photo_index, sizeof(photo_index), "%i", k);
	char coordinate[32];
	snprintf(coordinate, sizeof(coordinate), "%s,%s", images[k].lat, images[k].lon);
	char filename[128];
	snprintf(filename, sizeof(filename), "%s.jpg", images[k].key);
	// prepare POST body
	struct curl_httppost *formpost = NULL;
	struct curl_httppost *lastptr = NULL;
	curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "sequenceId", CURLFORM_PTRCONTENTS, openstreetcam_sequence_id, CURLFORM_END);
	curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "sequenceIndex", CURLFORM_PTRCONTENTS, photo_index, CURLFORM_END);
	curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "coordinate", CURLFORM_PTRCONTENTS, coordinate, CURLFORM_END);
	curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "headers", CURLFORM_PTRCONTENTS, images[k].heading, CURLFORM_END);
	curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "access_token", CURLFORM_PTRCONTENTS, openstreetcam_key, CURLFORM_END);
	curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "photo", CURLFORM_BUFFER, filename, CURLFORM_BUFFERPTR, images[k].data.c_str(), CURLFORM_BUFFERLENGTH, images[k].data.length(), CURLFORM_CONTENTTYPE, "image/jpeg", CURLFORM_END);
	// experimental fields
	if(images[k].captured_at[0]){
		curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "shotDate", CURLFORM_PTRCONTENTS, images[k].captured_at, CURLFORM_END);
	}
	bool is_sphere = (images[k].height > 0 && images[k].height * 2 == images[k].width);
	if(is_sphere){
		curl_formadd(&formpost, &lastptr, CURLFORM_COPYNAME, "projection", CURLFORM_COPYCONTENTS, "SPHERE", CURLFORM_END);
	}
	// curl options
	string data = "";
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &data);
	curl_easy_setopt(curl, CURLOPT_URL, "https://openstreetcam.org/1.0/photo/");
	curl_easy_setopt(curl, CURLOPT_HTTPPOST, formpost);
	double t1 = microtime();
	CURLcode res = curl_easy_perform(curl);
	double t2 = microtime();
	curl_formfree(formpost);
	if(res != CURLE_OK){
		log("[ul] Curl returned: %s\r\n", curl_easy_strerror(res));
		snprintf(last_error, sizeof(last_error), "Failed to upload image %i, curl error %i", k, res);
		return false;
	}
	long response_code;
	curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &response_code);
	if(response_code == 400 && strstr(data.c_str(), "You are not allowed to add a duplicate entry") != NULL){
		log("%s\r\n", data.c_str());
		log("[ul] Skipping duplicate image %i\r\n", k);
		return true;
	}
	if(response_code != 200){
		log("[ul] Server returned code: %li\r\n", response_code);
		log("[ul] Server said: %.1024s\r\n", data.c_str());
		snprintf(last_error, sizeof(last_error), "Failed to upload image %i, server returned code %li", k, response_code);
		return false;
	}
	log("%s\r\n", data.c_str());
	int len = images[k].data.length();
	progress_ul_speed = (int)((double)len/(t2-t1));
	progress_bytes_uploaded += len;
	log("[ul] %i/%i, %i bytes in %.3fs UL speed: %i bytes/s, sphere=%i\r\n", k+1, progress_images_discovered, len, (t2-t1), progress_ul_speed, (is_sphere ? 1 : 0));
	return true;
}

bool thread_ul_running = false;
void * thread_ul(void * arg){
	CURL * curl = curl_init_openstreetcam();
	if(!curl){
		thread_ul_running = false;
		return NULL;
	}

	// wait for at least one file to be downloaded
	while(thread_dl_running && !images[progress_images_uploaded].downloaded){
		usleep(250000); // wait for the file to be downloaded
	}
	if(!images[progress_images_uploaded].downloaded){
		log("[ul] Ending due to downloader failure\r\n");
		goto ul_end;
	}

	// create new sequence (if not provided)
	if(openstreetcam_sequence_id[0] != '\0'){
		log("[sequence] Continuing with sequence id %s\r\n", openstreetcam_sequence_id);
	}else{
		// create new sequence
		int i;
		bool created = false;
		for(i=0; i<10; i++){
			if(i > 0){
				log("[ul] Making another attempt to create new sequence...\r\n");
			}
			if(openstreetcam_new_sequence(curl)){
				created = true;
				break;
			}
			sleep(5);
		}
		if(!created){
			goto ul_end;
		}
		if(i > 0){
			last_error[0] = '\0'; // clear the error if recovered
		}
	}

	// start file uploading (start from progress_images_uploaded)
	int k;
	for(k=progress_images_uploaded; k<progress_images_discovered; k++){
		while(thread_dl_running && !images[k].downloaded){
			usleep(250000); // wait for the file to be downloaded
		}
		if(!images[k].downloaded){
			log("[ul] Ending due to downloader failure\r\n");
			break;
		}
		// upload file
		int i;
		bool uploaded = false;
		for(i=0; i<10; i++){
			if(i > 0){
				log("[ul] Making another attempt to upload...\r\n");
			}
			if(openstreetcam_upload(curl, k)){
				uploaded = true;
				break;
			}
			sleep(5);
		}
		if(!uploaded){
			break;
		}
		if(i > 0){
			last_error[0] = '\0'; // clear the error if recovered
		}
		// count
		progress_images_uploaded++;
		// free the memory
		images[k].data.~string();
	}

	// close the sequence
	if(progress_images_uploaded == progress_images_discovered){
		log("[ul] Closing sequence...\r\n");
		int i;
		bool closed = false;
		for(i=0; i<10; i++){
			if(i > 0){
				log("[ul] Making another attempt to close sequence...\r\n");
			}
			if(openstreetcam_close_sequence(curl)){
				closed = true;
				break;
			}
			sleep(5);
		}
		if(closed && i > 0){
			last_error[0] = '\0'; // clear the error if recovered
		}
	}

ul_end:
	log("[ul] Uploading thread finished, %i/%i images uploaded\r\n", progress_images_uploaded, progress_images_discovered);
	curl_easy_cleanup(curl);
	thread_ul_running = false;
	return NULL;
}

int main(int argc, char ** argv){
	if(argc < 6){
		printf("Usage: %s mapillary_client_id mapillary_key mapillary_sequence openstreetcam_key task_id [openstreetcam_sequence [resume_from]]\r\n", argv[0]);
		return 1;
	}
	mapillary_client_id = argv[1];
	mapillary_key = argv[2];
	mapillary_sequence = argv[3];
	openstreetcam_key = argv[4];
	task_id = argv[5];

	// init mutexes
	pthread_mutex_init(&printf_mutex, NULL);

	// open log file
	char log_fn[256];
	snprintf(log_fn, sizeof(log_fn), "log/mp2osc_%s.log", task_id);
	if(file_exists(log_fn)){
		printf("File %s already exists, can't continue\r\n", log_fn);
		return 1;
	}
	logfile = fopen(log_fn, "wb");
	if(!logfile){
		printf("Can't open log file %s for writing\r\n", log_fn);
	}

	// resuming capability (optional)
	if(argc > 6){
		strncpy(openstreetcam_sequence_id, argv[6], sizeof(openstreetcam_sequence_id));
	}
	if(argc > 7){
		int skip = atoi(argv[7]);
		progress_images_downloaded = skip;
		progress_images_uploaded = skip;
		log("Skipping %i files\r\n", skip);
	}

	// init curl
	curl_global_init(CURL_GLOBAL_DEFAULT);
	libcurl_version = curl_version_info(CURLVERSION_NOW);

	// report that the worker is started
	CURL * curl_backend = curl_init_backend();
	bool start_reported = false;
	int i;
	for(i=0; i<5; i++){
		if(progress(curl_backend, false, false, NULL)){
			start_reported = true;
			break;
		}
		sleep(5);
	}
	if(!start_reported){
		// If the backend doesn't know that we have started, then it can run the same job again, which will end in duplicates
		// Don't do anything if we can't contact the backend
		log("Initial contact with backend failed\r\n");
		return 1;
	}

	// get mapillary sequence details
	CURL * curl_mapillary = curl_init_mapillary();
	bool got = false;
	for(i=0; i<5; i++){
		if(i > 0){
			log("[sequence] Making another attempt to get sequence details...\r\n");
		}
		if(get_mapillary_image_list(curl_mapillary)){
			got = true;
			break;
		}
		progress(curl_backend, false, false, (last_error[0] ? last_error : "Failed to get image list, retrying"));
		sleep(5);
	}
	if(!got){
		progress(curl_backend, true, false, (last_error[0] ? last_error : "Failed to get image list"));
		return 1;
	}else if(i > 0){
		last_error[0] = '\0'; // clear the error if recovered
	}
	progress(curl_backend, false, false, last_error);

	// start downloading thread
	pthread_t dl_thread;
	thread_dl_running = true;
	pthread_create_detached(&dl_thread, thread_dl, curl_mapillary);

	// start uploading thread
	pthread_t ul_thread;
	thread_ul_running = true;
	pthread_create_detached(&ul_thread, thread_ul, NULL);

	// main loop
	i = 0;
	while(1){
		if(!thread_dl_running && progress_images_downloaded != progress_images_discovered){
			log("[main_loop] Downloading failed\r\n");
			// wait up to 20s for the uploading thread to finish
			int i;
			for(i=0; i<20; i++){
				if(!thread_ul_running){
					break;
				}
				sleep(1);
			}
			break;
		}
		if(!thread_ul_running){
			log("[main_loop] Uploading %s\r\n", (progress_images_uploaded == progress_images_discovered ? "finished" : "failed"));
			break;
		}
		usleep(250000);
		if(++i % 4 == 0){
			progress(curl_backend, false, false, last_error);
		}
	}

	// report that the worker has finished
	bool success = (progress_images_downloaded == progress_images_discovered && progress_images_uploaded == progress_images_discovered);
	if(!success && !last_error[0]){
		snprintf(last_error, sizeof(last_error), "Unknown error");
	}
	for(i=0; i<60; i++){
		if(progress(curl_backend, true, success, last_error)){
			break;
		}
		sleep(5);
	}

	// cleanup
	curl_easy_cleanup(curl_mapillary);
	curl_easy_cleanup(curl_backend);
	curl_global_cleanup();
	log("Program finished\r\n");
	return 0;
}
